////////////////////////////////////////////////////////////////////////
/*! \file   lambert_fragment.glsl
 *  \brief  Lambert conform projection. Common fragment shader.
 *  \author J.D. Gascuel
 *  \sa common_vertex.glsl, lambertTris_geometry.glsl
 */
////////////////////////////////////////////////////////////////////////
// $Id: lambert_fragment.glsl,v 1.11 2007/12/08 18:47:49 gascuel Exp $
// File History:
// [jDG 2006/12/05] : Creation from Cg version
// [jDG 2007/12/04] : Added No fragment mode (I3D2008 statistics)

#version 120
#extension EXT_texture_array    : enable

const float epsilon0 = (1.0/512.0)/20;  // 1/20 pixel size (in wp coords)
const float epsilon1 =  1.0 + 2*epsilon0;

//----------------------------------------------------------------------------
// Per Scene data :

uniform float     uFar = 1.0;       // Scene far distance.

// Lighting. If 0, use interpolated per-vertex-colors.
uniform int uLights;

// Color texture, if any.
uniform bool      uHasColorMap;
uniform sampler2D uColorMap;

// Additional environment mapping, if any.
#if defined(PARABOLIC_ENV_MAP)
    uniform bool uHasParabolicEnvMap;
#   ifdef PARABOLIC_ENV_MAP_ARRAY
        uniform sampler2DArray uEnvMapParabolic;
#   else
        uniform sampler2D uEnvMapParabolic;
#   endif
#elif defined(ENV_MAP)
    uniform bool uHasEnvMap;
    uniform samplerCube uEnvMap;
#endif

// Bump texture, if any.
#ifdef BUMP_MAPPING
uniform bool uHasBumpMap;
uniform sampler2D uBumpMap;
#endif

// Shadow map data :
#ifdef SHADOW_MAPPING
uniform bool uHasShadow;			// Should we apply shadow map to this object ?
uniform sampler2DShadow uShadowMap;	// Projection from world to light frame.
#endif

// Overdraw statistics: add do-nothing pass :
uniform bool uTestOverdraw = false;

// Per triangle data :
flat varying in vec3    oColors[4]; // RGBA components, so that RGBA.i = dot(coord.xyz, oColors[i])
flat varying in vec3    oU, oV;     // If uHasTexture is true, tex coordinate U = dot(coord.xyz, oUV[0])
flat varying in vec3    oNp[3];     // For per-pixel-lighting, N.i = dot(coord.xyz, oN[i]))
flat varying in vec3	oSp[4];		// Position in light's frame.

// Per pixel data :
varying in vec2 oP;                 // Raster position in 3D space.

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

vec3
unProject(vec2 P)
{
    // Inverse the lens function to recover the incoming ray direction.
    float a = dot(oP.xy, oP.xy);                    // 1 - Z/rho

    if( a <= 1.0f )          // Are we inside the hemispherical viewport ?
    {
        // Inside the viewport : do the full reconstruction (with z < 0).
        float cosTheta = 1.0 - a;                       // Z/rho
        float sinTheta = sqrt(1.0 - cosTheta*cosTheta); // Theta in 0..pi/2
        return vec3( oP * sinTheta/sqrt(a), -cosTheta);
    }
 
    return vec3(0,0,1);     // z > 0 mark outside viewport...
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#if !defined(ISECT_BADOUEL) && !defined(ISECT_MTH)
#	define	ISECT_BADOUEL
#endif

#ifdef ISECT_BADOUEL

//////////////////////////////////////////////////////////////////////////////
/// Compress vec3 to vec2 by skipping coordinates designated by dominant.
vec2
crunsh(in vec3 v, in float dominant)
{
    if( dominant <  0.0 ) return v.yz;
    if( dominant == 0.0 ) return v.xz;
                          return v.xy;
}

////////////////////////////////////////////////////////////////////////
/// Badouel ray-triangle intersection.
///
/// Returns (1,alpha,beta) such that P = A + alpha (B-A) + beta (C-A)
/// If P is not within the triangle (A,B,C), discard the fragment.

// Per-triangle intersection data :
flat varying float   oDominant; 
flat varying vec3    oA;        // Original triangle world coordinates, projected on the
flat varying vec2    oAB;       // unit 3D sphere.
flat varying vec2    oAC;
flat varying vec3    oN;        // Triangle's unit normal vector.

vec4
intersect(in vec3 P)
{
    // d is on the unit sphere. We need to project it onto the triangle's
    // plane (A, AB, AC). That is find P' = alpha P such that :
    //  AP' . oN = 0
    //   P' . oN = oA . oN
    // alpha = dot(A, oN) / dot(P, oN)
    float depth = dot(oN, oA) / dot(oN, P);
    if( depth < 0.0 )       // Is the wrong side hiting the triangle ?
    {
#ifndef SUBMODE_DEBUG
        if( uTestOverdraw )
#endif
		    return vec4(-1.0, 0, 0, -1);
        discard;
    }
    P *= depth;

    vec2 AP = crunsh(P-oA, oDominant);

    // Solve the 2 eq. linear system :
    //   AP.x = alpha oAB.x + beta oAC.x
    //   AP.y = alpha oAB.y + beta oAC.y
    //
    //float det =  oAB.x * oAC.y - oAB.y * oAC.x;
    //    alpha = (AP.x * oAC.y - AP.y * oAC.x) / det;
    //    beta  = (oAB.x * AP.y - oAB.y * AP.x) / det;
    //
    // Optimize by using a dot2(), instead of mul+madd :
    vec2 oACt = vec2(oAC.y, -oAC.x);
    vec2 oABt = vec2(-oAB.y, oAB.x);
    float det = dot(oAB, oACt);

    vec4 coord = vec4(1.0, dot(AP, oACt), dot(oABt, AP), depth);
    if( det != 0.0 )
    {
        coord.yz /= det;
    
        // If we are inside the triangle, skip the final discard.
        if( (coord.y >= -epsilon0) && (coord.z >= -epsilon0) && ((coord.y+coord.z) <= epsilon1) )
            return coord;
    }

    // ---- Kill pixels outside the original triangle :
#ifndef SUBMODE_DEBUG
    if( uTestOverdraw )
#endif
        return vec4(-1.0,0,0,1);  // Mark outside, just for debug purpose.
    discard;
}

#endif // ISECT_BADOUEL

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

#ifdef ISECT_MTH

// Per-triangle intersection data :
flat varying vec3    oN;        // AB cross AC
flat varying vec3    oQ1;		// AO cross AB
flat varying vec3    oQ2;       // AO cross AC
flat varying float   oT;		// Q1 dot AC

vec4
intersect(const in vec3 P)
{
	float det = - dot(P, oN);
	if( det == 0.0 )
	{
#ifndef SUBMODE_DEBUG
        if( uTestOverdraw )
#endif
            return vec4(-1.0,0,0,1);  // Mark outside, just for debug purpose.
        discard;
    }
	vec4 coord = vec4(1.0, - dot(P, oQ2), dot(P, oQ1), oT);
	coord.yzw *= 1.0 / det;

    // If we are inside the triangle, skip the final discard.
    if( (coord.y >= -epsilon0) && (coord.z >= -epsilon0) && ((coord.y+coord.z) <= epsilon1) )
        return coord;

#ifndef SUBMODE_DEBUG
    if( uTestOverdraw )
#endif
        return vec4(-1.0,0,0,1);  // Mark outside, just for debug purpose.
    discard;
}

#endif // ISECT_MTH

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

///
/// Directional-light calculation.
/// Derived from OpenGL Shading Language, second edition, p 237.

void
directionalLight(int  i,         //!< Light number.
                 vec3 E,         //!< Eye direction (usually +z)
                 vec3 N,         //!< normal vector, in camera coordinates.
                 inout vec4 color)
{
    vec3 V = normalize(gl_LightSource[i].position.xyz); // Light direction.
    vec3 H = normalize(V + E);  // mid between surface-to-light and look direction.

    float nDotV = max(0.0, dot(N, V));
    float nDotH = max(0.0, dot(N, H));

    float pf = 0.0;
    if( nDotV > 0.0 && gl_FrontMaterial.shininess > 0.0 )   // Shinny only on the front side.
        pf = pow(nDotH, gl_FrontMaterial.shininess);

    color  += gl_FrontLightProduct[i].ambient
           +  gl_FrontLightProduct[i].diffuse  * nDotV
           +  gl_FrontLightProduct[i].specular * pf;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Point-light calculation.
/// Derived from OpenGL Shading Language, second edition, p 238.

void
pointLight(int  i,         //!< Light number.
           vec3 P,         //!< pixel position, in camera coordinates.
           vec3 E,         //!< Eye direction (usually +z)
           vec3 N,         //!< normal vector, in camera coordinates.
           inout vec4 color)
{
    vec3 V =  gl_LightSource[i].position.xyz - P;   // Surface to ligth vector.
    float d = length(V);        // Distance from light source.
    V = normalize(V);           // Light direction (in camera coordinates).

    float atn = 1.0 / ( gl_LightSource[i].constantAttenuation
                      + gl_LightSource[i].linearAttenuation    * d
                      + gl_LightSource[i].quadraticAttenuation * d * d);

    vec3 H = normalize(V + E);  // mid between surface-to-light and look direction.
    float nDotV = max(0.0, dot(N, V));
    float nDotH = max(0.0, dot(N, H));

    float pf = 0.0;
    if( nDotV > 0.0 && gl_FrontMaterial.shininess > 0.0 ) // Shinny only on the front side.
        pf = pow(nDotH, gl_FrontMaterial.shininess);

    color  += gl_FrontLightProduct[i].ambient
           +  gl_FrontLightProduct[i].diffuse  * nDotV * atn
           +  gl_FrontLightProduct[i].specular * pf * atn;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void main()
{
    // oP is the original 2D vertex coordinate (the only thing we are
    // allowed to interpolate).
    vec3 d = unProject(oP);
    
    if( d.z > 0.0 )    // Marked outside viewport ?
    {
#ifdef SUBMODE_DEBUG
        gl_FragDepth = 0.9999;                      // forced at infinity.
        gl_FragColor = vec4(0,0,0.6,0.05);			// Dark-blue outside viewport.
        return;
#else
        discard;
#endif
    }
    
    // ---- Compute baryrentric coordinates ----------------------------------
    // coord.x == 1.0 is inside, -1.0 if outside.
    // coord.y == alpha
    // coord.z == beta
    // coord.w == depth multiplier.
    vec4 coord = intersect(d);

    // glDepthRange() data is clipped to [0..1], so useless here...
    float depth = coord.w / uFar;

#ifdef SUBMODE_COVERAGE //////////////////////////////////////////////////////

	gl_FragColor = vec4(0.5,0.5,0.5, 1);
    gl_FragDepth = 0.999;     // forced at infinity.

#elif defined(SUBMODE_DEPTH) /////////////////////////////////////////////////

    gl_FragColor = vec4(depth, depth, depth, 1);
    gl_FragDepth = depth;

#elif defined(SUBMODE_DEBUG) /////////////////////////////////////////////////

    // DEBUG: color.rg = alpha,beta, color.b=1 when outside.
    if( coord.x >= 0.0 )    // sould be +1 if inside triangle
        gl_FragColor = vec4(coord.yz, 0, 0.50);
    else if( coord.w > 0.0 )	// Front side ?
        gl_FragColor = vec4(0,0,1, 0.05);
    else						// Back side.
        gl_FragColor = vec4(0,0,0.5, 0.05);
    gl_FragDepth = 0.999;     // forced at infinity.

#else // SUBMODE_COLOR ///////////////////////////////////////////////////////

    // If the pixel is outside the triangle, 
    // badouel() already have discarded it...
    
    // We know the distance in real word coordinates : juste remap-it.
    gl_FragDepth = depth;

#ifdef COLOR_TEXTURE
    vec2 tcoord = vec2( dot(coord.xyz, oU),
                        dot(coord.xyz, oV));
#endif

#ifdef TEST_SHADING //////////////////////////////////////////////////////////

    // Fake shading: use parameters (to avoid over-optimisations),
    // but don't do any time-consumming computations...
    
    // Interpolat RGBA coordinates
#ifdef PER_PIXEL_LIGHTINGS
    gl_FragColor = gl_FrontLightModelProduct.sceneColor
                 + gl_FrontLightProduct[0].ambient
                 + gl_FrontLightProduct[0].diffuse;

    // Interpolat Normal coordinates.
    gl_FragColor.r *= 0.5 + 0.5 * dot(coord.xyz, oNp[0]);
    gl_FragColor.g *= 0.5 + 0.5 * dot(coord.xyz, oNp[1]);
    gl_FragColor.b *= 0.5 + 0.5 * dot(coord.xyz, oNp[2]);
#   ifdef SHADOW_MAPPING
    // If needed, interpolate shadow coordinates.
    if( uHasShadow ) // per object shadow-mapping enable.
    {
	    vec4 oS = vec4( dot(coord.xyz, oSp[0]),
	                    dot(coord.xyz, oSp[1]),
	                    dot(coord.xyz, oSp[2]),
	                    dot(coord.xyz, oSp[3]) );
        gl_FragColor.rgb *= oS.xyz / oS.w;
    }
#   endif
#else // Interpolated colors...
    gl_FragColor.r = dot(coord.xyz, oColors[0]);
    gl_FragColor.g = dot(coord.xyz, oColors[1]);
    gl_FragColor.b = dot(coord.xyz, oColors[2]);
    gl_FragColor.a = dot(coord.xyz, oColors[3]);
#endif

#ifdef COLOR_TEXTURE
    // If needed, use interpolat texture coordinates.
    if( uHasColorMap )
    {
        gl_FragColor.r *= tcoord.r;
        gl_FragColor.g *= tcoord.s;
    }
#endif

#else // TEST_SHADING ////////////////////////////////////////////////////////

#if defined(PER_PIXEL_LIGHTINGS) || defined(PARABOLIC_ENV_MAP) || defined(ENV_MAP)
    // Material property lighting : p 241 :
    vec3 P = d * coord.w;

    vec3 normal;
#if defined(BUMP_MAPPING) && defined(COLOR_TEXTURE)
    if( uHasBumpMap )
    {
	    // TODO : should go to tangeant space here !
	    normal = normalize(texture2D(uBumpMap, tcoord).xyz);
    }
    else
#endif
    normal = normalize( vec3( dot(coord.xyz, oNp[0]),
                              dot(coord.xyz, oNp[1]),
                              dot(coord.xyz, oNp[2]) ));

#if defined(PER_PIXEL_LIGHTINGS)
    // Viewing direction is fragment depedent.
    vec3 E = -normalize(P);

#ifdef SHADOW_MAPPING
    // TODO : percentage closer filtering:
    //   1- shadow --> float.
    //   2- have space-sized jittering. That is before w division ?
    //   3- change oS too. Add dS/dU * dU ?
    //   4- make N lookups. average all comparaison results.
    bool shadowed = false;
    if( uHasShadow ) // per object shadow-mapping enable.
    {
	    vec4 oS = vec4( dot(coord.xyz, oSp[0]),
	                    dot(coord.xyz, oSp[1]),
	                    dot(coord.xyz, oSp[2]),
	                    dot(coord.xyz, oSp[3]) );
        shadowed = oS.w * shadow2DProj(uShadowMap, oS).z <= oS.z;
    }
#endif

    gl_FragColor = gl_FrontLightModelProduct.sceneColor;	// emissive + global ambient light.
    for(int l=0; l < uLights; ++l)
    {
#ifdef SHADOW_MAPPING
	    if( l==0 && shadowed )
		    gl_FragColor += gl_FrontLightProduct[0].ambient;
	    else
#endif
	    {
        if( gl_LightSource[l].position.w == 0.0 )
			    directionalLight(l, E, normal, gl_FragColor);
        else    // Let's suppose it is not a spot (ie. spotCutoff == 180.0)
			    pointLight(l, P, E, normal, gl_FragColor);
	    }
    }
#else // PARABOLIC_ENV_MAP, but no PER_PIXEL_LIGHTINGS
    gl_FragColor.r = dot(coord.xyz, oColors[0]);
    gl_FragColor.g = dot(coord.xyz, oColors[1]);
    gl_FragColor.b = dot(coord.xyz, oColors[2]);
    gl_FragColor.a = dot(coord.xyz, oColors[3]);
#endif 
#else // no PER_PIXEL_LIGHTINGS and no PARABOLIC_ENV_MAP
    // Any other parameters are just interpolated from barycentric coords.
    gl_FragColor.r = dot(coord.xyz, oColors[0]);
    gl_FragColor.g = dot(coord.xyz, oColors[1]);
    gl_FragColor.b = dot(coord.xyz, oColors[2]);
    gl_FragColor.a = dot(coord.xyz, oColors[3]);
#endif

#ifdef COLOR_TEXTURE
    if( uHasColorMap )
        gl_FragColor.xyz *= texture2D( uColorMap, tcoord).xyz;
#endif // COLOR_TEXTURE

#if defined(PARABOLIC_ENV_MAP)
    if( uHasParabolicEnvMap )
    {
        vec3 r = reflect(normalize(P), normal);         // In camera's coordinates.
        r = mat3(gl_ModelViewMatrixTranspose) * r;      // Back to world orientation.

        // The paraboloid transform, for a unit vector:
        r /= 1.0 + abs(r.z);

#   ifdef PARABOLIC_ENV_MAP_ARRAY
        // Use 2 layers of a 2D texture array : map to [0..1]x[0..1]x{0,1}
        if( r.z < 0.0 )
            r = vec3(0.5, 0.5, 1.0) + vec3(-0.49,-0.498, 0.0) * r;
        else
            r = vec3(0.5, 0.5, 0.0) + vec3(-0.49, 0.498, 0.0) * r;

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.xyz = mix(gl_FragColor.xyz, texture2DArray(uEnvMapParabolic, r).xyz, 0.5);
#   else
        // Use a single texture, twice as height : map to [0..1]x[0..2/2]
        if( r.z < 0.0 )
            r.xy = vec2(0.5, 0.75) + vec2(-0.49,-0.248) * r.xy;
        else
            r.xy = vec2(0.5, 0.25) + vec2(-0.49, 0.248) * r.xy;

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.xyz = mix(gl_FragColor.xyz, texture2D(uEnvMapParabolic, r.xy).xyz, 0.5);
#   endif
    }
#elif defined(ENV_MAP)
    if( uHasEnvMap )
    {
        vec3 r = reflect(normalize(P), normal);         // In camera's coordinates.
        r = mat3(gl_ModelViewMatrixTranspose) * r;      // Back to world orientation.

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.xyz = mix(gl_FragColor.xyz, textureCube(uEnvMap, r).xyz, 0.5);
    }
#endif
    // Keep alpha channel safe...
    gl_FragColor.a = gl_Color.a;

#endif  // TEST_SHADING ////////////////////////////////////////////////////////

#endif // SUBMODE_COLOR ////////////////////////////////////////////////////////
}