////////////////////////////////////////////////////////////////////////
/*! \file   sphereTris_geometry.glsl
 *  \brief  Cosinus-Sphere projection. Trinagles flavor.
 *  \author J.D. Gascuel
 *  \sa common_vertex.glsl, sphere_fragment.glsl
 */
////////////////////////////////////////////////////////////////////////
// $Id: sphereTris_geometry.glsl,v 1.42 2007/12/08 12:50:14 gascuel Exp $
// File History:
// [jDG 2006/12/05] : Creation from Cg version

#version 120
#extension EXT_geometry_shader4 : require
#extension EXT_gpu_shader4      : require

//----------------------------------------------------------------------------

const float epsilon0  = 0.0001f;
const float epsilon1  = 1 + epsilon0;
const float epsilonm1 = 1 - epsilon0;

//----------------------------------------------------------------------------

// Per rendering data :
uniform int  uPrimitiveFilter = -1; // Retain only one triangle if >= 0.
uniform int  uCulling = 1;			// Retain only front facing triangles.

uniform bool uHasBumpMap = false;
uniform bool uHasColorMap = false;

flat varying out vec3[4] oColors;   // RGBA components, so that RGBA.i = dot(coord, oColors[i])
flat varying out vec3    oU, oV;    // If uHasTexture is true, tex coordinate U = dot(coord, oU)

// Per pixel data :
varying out vec2 oP;  // Raster position in 3D space.

// For per-pixel-lighting:
varying in  vec3        iN[];       // Normal, output from vertex shader
flat varying out vec3   oNp[3];     // Normal interpolation data (for per-pixel-lighting).

// Shadow map data :
varying in  vec4 iS[];				// Position, in light[1] frame.
flat varying out vec3 oSp[4];		// light space interpolation data.

////////////////////////////////////////////////////////////////////////

#define EMIT_VERTEX(v2D)                    \
{                                           \
    oP = v2D;                               \
    gl_Position = vec4(v2D, 0, 1);          \
    EmitVertex();                           \
}

//----------------------------------------------------------------------------
/// Projection functions. From a eye-space point, to the 2D screen space.

float
radialFunction(vec3 P)
{
    return length(P);
}

vec3
projection(vec3 P)
{
    return normalize(P);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// Where is the maximum deformation over an arc on the projected sphere ?
/// \p backFace should be < for backface triangles.

vec2
maxDeformation(vec3 A, vec3 B, float backFace)
{
    // Edge deformation is always centrifuge.
    // But, does the inside of the triangle is toward
    // O or not ?

    vec3 C = B.xyz-A.xyz;
    float inward = backFace * (A.x * C.y - A.y * C.x);
    if( inward > 0 )
    {
        //---- Max is at the midpoint of the 2D arc !
        // Note: becase we had projected to the unit sphere, the
        //       mid-point is always INSIDE the sphere, hence 
        //       result is always > 1.0
        C = A+B; // Midpoint, with an arbitrary multiplicator.
        return epsilon1 * projection(C).xy;
    }

    return epsilonm1 * A.xy;
}

//----------------------------------------------------------------------------

/// 2D line intersection.
///
/// Solve in (alpha,beta)
///     A + alpha U = B + beta V
/// That is :
///   (B-A).x = alpha U.x - beta V.x
///   (B-A).y = alpha U.y - beta V.y
///
vec2
intersect(vec2 A, vec2 U,
          vec2 B, vec2 V)
{
    float d = U.y*V.x - U.x*V.y;

    if( d != 0.0 )      // Oups, U and V parallel... Just keep A vertex.
    {
        B -= A;
        d = (V.x * B.y - V.y * B.x) / d;
        A += d * U;
    }
    return A;
}

//----------------------------------------------------------------------------
/// \brief Compute a bigger 2D triangle, from the 3 3D projected vertices.
void
convexTriangle(vec3 A,
               vec3 B,
               vec3 C,
               float backFace)
{
    vec2 AB = B.xy - A.xy;
    vec2 BC = C.xy - B.xy;
    vec2 CA = A.xy - C.xy;

    // Enlarge each vertice of the triangle, acording to the arc deformation
    // of the projected edges :
    vec2 Ap = maxDeformation(A, B, backFace);
    vec2 Bp = maxDeformation(B, C, backFace);
    vec2 Cp = maxDeformation(C, A, backFace);

    // And compte the new intersections :
    vec2 App = intersect(Ap, AB, Bp, BC);      // Beware AB is not Bp-Ap ...
    vec2 Bpp = intersect(Bp, BC, Cp, CA);
    vec2 Cpp = intersect(Cp, CA, Ap, AB);
    
    // Then emit the bigger triangle coordinates :
    EMIT_VERTEX(App)
    EMIT_VERTEX(Bpp)
    EMIT_VERTEX(Cpp)
    EndPrimitive();
}

//----------------------------------------------------------------------------
/// \brief Compute a projected bigger 2D triangle, from the 3 3D vertices.
void
zClipedTriangle(vec3 iA,
                vec3 iB,
                vec3 iC,
                const in float backFace)
{
    // Special projection on the 3D unit sphere :
    vec3 A = projection(iA);
    vec3 B = projection(iB);
    vec3 C = projection(iC);
 
    // Compute edge direction, on the projected plane :
    vec2 AB = B.xy - A.xy;
    vec2 BC = C.xy - B.xy;
    vec2 CA = A.xy - C.xy;
    
    // ---- Detect obtuse triangles ------------------------------------------
    // If any of the angle is negativ, then all three are.
    // Hence, just test angle(AB, BC) :
    float a = (AB.x * BC.y - AB.y * BC.x) * backFace;
    if( a < -epsilon0 )
    {
        // We should cut the longer edge of the projected triangle :
        float ab = length(AB);
        float bc = length(BC);
        float ca = length(CA);
        
        // Rotate to have longer edge in AC ?
        if( (bc > ca) && (bc >= ab) )
        {
            vec3 tmp =  A;  A =  C;  C =  B;  B = tmp;
                 tmp = iA; iA = iC; iC = iB; iB = tmp;
        }       
        else if( (ab > ca) )    // Rotate-left to have longer edge in AC ?
        {
            vec3 tmp =  A;  A =  B;  B =  C;  C = tmp;
                 tmp = iA; iA = iB; iB = iC; iC = tmp;
        }       

        //vec2 Bt = vec2(-iB.y, iB.x);
        //float lambda = dot(iA.xy, Bt)/dot(iA.xy - iC.xy, Bt);
        float lambda = ( iA.y*iB.x - iB.y*iA.x );
              lambda = lambda / (lambda  - iC.y * iB.x + iB.y * iC.x ); 
        vec3 D = projection(mix(iA, iC, lambda));
        
        convexTriangle(A, B, D, backFace);
        A = D;
    }

    // ---- Back to general case ---------------------------------------------
        convexTriangle(A, B, C, backFace);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#if !defined(ISECT_BADOUEL) && !defined(ISECT_MTH)
#	define	ISECT_BADOUEL
#endif

#ifdef ISECT_BADOUEL

//////////////////////////////////////////////////////////////////////////////
/// Which is the biggest componant of a vector (absolute value).
/// Returns -1 for X, 0 for Y, and +1 for Z (best for if()).
float
dominant(in vec3 v)
{
    v = abs(v);
    return ( v.x >= v.y && v.x >= v.z ) ? -1.0
         : ( v.y >= v.x && v.y >= v.z ) ?  0.0
         :                                 1.0;
}

//////////////////////////////////////////////////////////////////////////////
/// Compress vec3 to vec2 by skipping coordinates designated by dominant.
vec2
crunsh(in vec3 v, in float dominant)
{
    if( dominant <  0.0 ) return v.yz;
    if( dominant == 0.0 ) return v.xz;
                          return v.xy;
}

//////////////////////////////////////////////////////////////////////////////
///

// Per-triangle intersection data :
flat varying float   oDominant; 
flat varying vec3    oA;        // Original triangle world coordinates, projected on the
flat varying vec2    oAB;       // unit 3D sphere.
flat varying vec2    oAC;
flat varying vec3    oN;        // Triangle's unit normal vector.

void
intersectSetup(vec3 A, vec3 B, vec3 C)
{
	B -= A;
	C -= A;

    oN         = cross(B, C);
    float d = dominant( oN );

    oDominant  = d;
    oA         = A;                 // Keep (oA, N) in 3D, to enable plane projection.
    oAB        = crunsh(B, d);
    oAC        = crunsh(C, d);
}

#endif // ISECT_BADOUEL

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#ifdef ISECT_MTH

// Per-triangle intersection data :
flat varying vec3    oN;        // AB cross AC
flat varying vec3    oQ1;		// AO cross AB
flat varying vec3    oQ2;       // AO cross AC
flat varying float   oT;		// Q1 dot AC

void
intersectSetup(vec3 A, vec3 B, vec3 C)
{
	B -= A;
	C -= A;
	oN  = cross(B, C);	// AB ^ AC
	oQ1 = cross(B, A);	// AB^A == AO^AB
	oQ2 = cross(C, A);	// AC^A == AO^AC
	oT  = dot(oQ1, C);	// C = AC
}

#endif // ISECT_MTH

//////////////////////////////////////////////////////////////////////////////

void main()
{
    // Filter primitive option :
#ifdef PRIMITIVE_FILTERING
    if( (uPrimitiveFilter >= 0) && ( gl_PrimitiveIDIn != uPrimitiveFilter) )
        return;
#endif

    vec3 A = gl_PositionIn[0].xyz;
    vec3 B = gl_PositionIn[1].xyz;
    vec3 C = gl_PositionIn[2].xyz;

    // ---- Is this triangle fully on the wrong side of the camera plane ?
    int code =     int( A.z > 0.0 )
             + 2 * int( B.z > 0.0 )
             + 4 * int( C.z > 0.0 );
    if( code == 7 ) return;

    // TwoPass test :
    // Keep only triangles that won't produce a quad by z-clipping:
    // if( code*(code-2)*(code-5)*(code-6) != 0 ) return;

    // ---- Or back-facing ? -------------------------------------------------
    // Triangles front-facing the eye are those one so that 
    //    dot(view, N) < 0
    // Because view = A + alpha AB + beta AC (point on plane),
    // and dot(N,AB) == dot(N,AC) == 0, it is sufficient to test
    // just any corner of the triangle :
    vec3 N  = cross(B-A, C-A);
    float backFace = - dot(A,N);
#ifdef FACE_CULLING   
    if( (uCulling * backFace) < 0.0 ) return;
#endif

	// ---- Setup attributes shared by all generated triangles ---------------
    intersectSetup(A, B, C);

    // ---- Prepare all attributes too ---------------------------------------
#ifdef SUBMODE_COLOR
#ifdef COLOR_TEXTURE
    if( uHasColorMap || uHasBumpMap )
    {
        vec4 d = vec4(gl_TexCoordIn[1][COLOR_TEXTURE].xy, 
                      gl_TexCoordIn[2][COLOR_TEXTURE].xy)
              - gl_TexCoordIn[0][COLOR_TEXTURE].xyxy;
        oU = vec3(gl_TexCoordIn[0][COLOR_TEXTURE].x, d.x, d.z);
        oV = vec3(gl_TexCoordIn[0][COLOR_TEXTURE].y, d.y, d.w);
    }
#endif

#ifdef PER_PIXEL_LIGHTINGS
    oNp[0] = vec3(iN[0].x, iN[1].x - iN[0].x, iN[2].x - iN[0].x);
    oNp[1] = vec3(iN[0].y, iN[1].y - iN[0].y, iN[2].y - iN[0].y);
    oNp[2] = vec3(iN[0].z, iN[1].z - iN[0].z, iN[2].z - iN[0].z);

#ifdef SHADOW_MAPPING
    oSp[0] = vec3(iS[0].x, iS[1].x - iS[0].x, iS[2].x - iS[0].x);
    oSp[1] = vec3(iS[0].y, iS[1].y - iS[0].y, iS[2].y - iS[0].y);
    oSp[2] = vec3(iS[0].z, iS[1].z - iS[0].z, iS[2].z - iS[0].z);
    oSp[3] = vec3(iS[0].w, iS[1].w - iS[0].w, iS[2].w - iS[0].w);
#endif

#else // no PER_PIXEL_LIGHTINGS
    oColors[0] = vec3(gl_FrontColorIn[0].r, gl_FrontColorIn[1].r - gl_FrontColorIn[0].r, gl_FrontColorIn[2].r - gl_FrontColorIn[0].r);
    oColors[1] = vec3(gl_FrontColorIn[0].g, gl_FrontColorIn[1].g - gl_FrontColorIn[0].g, gl_FrontColorIn[2].g - gl_FrontColorIn[0].g);
    oColors[2] = vec3(gl_FrontColorIn[0].b, gl_FrontColorIn[1].b - gl_FrontColorIn[0].b, gl_FrontColorIn[2].b - gl_FrontColorIn[0].b);
    oColors[3] = vec3(gl_FrontColorIn[0].a, gl_FrontColorIn[1].a - gl_FrontColorIn[0].a, gl_FrontColorIn[2].a - gl_FrontColorIn[0].a);
#endif // PER_PIXEL_LIGHTINGS
#endif // SUBCOLOR_MODE

    // ---- Cover the bended triangle with some rendering area -----------
    //
    // Make everybody ~ case 1 : cut A, keep B & C,
    //          or like case 6 : keep corner A, cut B & C.
    if( (code-3) * (code-4) == 0 ) { vec3 tmp = A; A = C; C = B; B = tmp; }
    if( (code-2) * (code-5) == 0 ) { vec3 tmp = A; A = B; B = C; C = tmp; }

    if( code != 0 )
    {
        // Exactly two edge intersection do exists.
        // Because of the previous permuttations, we do know that they are
        // indeed intersections on the A-B and the A-C edges :
        vec3 Bp = mix(A, B, A.z/(A.z-B.z));
        vec3 Cp = mix(A, C, A.z/(A.z-C.z));

        if( (code-1) * (code-2) * (code-4) == 0 )               // Just one corner clipped ?
        {
            zClipedTriangle(Bp, C, Cp, backFace);
            A = Bp;
        }
        else if( (code-3) * (code-5) * (code-6) == 0 )          // Exactly two corner clipped ?
        {
            B = Bp;
            C = Cp;
        }
    }

    // Note: code 0 (no clip) is ~ keeping one triangle, but with keeping B anc C
    //       values.
    zClipedTriangle(A, B, C,  backFace);
}
